// -*- c++ -*-

#include <qevent.h>
#include <qkeycode.h>
#include <qheader.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qclipboard.h>

#include <kapplication.h>
#include <kfileitem.h>
#include <kglobal.h>
#include <kglobalsettings.h>
#include <kicontheme.h>
#include <kdebug.h>
#include <qdragobject.h>
#include <kaction.h>
#include <kiconloader.h>
#include <ktrader.h>
#include <krun.h>

#include "pfileview.h"
#include "pfilepanel.h"
#include "pcmm_general.h"

#define NCOLS 6
#define COL_NAME 0
#define COL_SIZE 1
#define COL_DATE 2
#define COL_PERM 3
#define COL_OWNER 4
#define COL_GROUP 5

#define UTF8_MIB 106

class PFileView::PFileViewPrivate {
public:
    PFileViewPrivate( bool left )
    : is_left_panel(left){};

    QStringList files_dropped_list;
    KURL dropped_to_url;

    bool m_blockSortingSignal;

    bool show_col_size;
    bool show_col_date;
    bool show_col_perm;
    bool show_col_owner;
    bool show_col_group;

    bool is_left_panel;

    KTrader::OfferList file_offers;
    KPopupMenu *open_popup;
};

PFileView::PFileView( bool left, QWidget *parent, const char *name )
: PListView(parent, name), cols( new int[NCOLS] ), d( new PFileViewPrivate( left ) ) {
    readConfig();
    
    //setItemsRenameable( true );
    setDefaultRenameAction( Reject );
    setItemsMovable( false );

    // Drag&Drop
    setDragEnabled( true );
    setAcceptDrops( true );
    setDropVisualizer(false);
    setDropHighlighter(true);
    connect( this, SIGNAL( dropped (QDropEvent *, QListViewItem *) ), SLOT( filesDropped( QDropEvent *) ) ); 

    m_sortingCol = cols[COL_NAME];
    d->m_blockSortingSignal = false;
    setViewName( i18n("Detailed View") );

    setShowSortIndicator( TRUE );
//	setAllColumnsShowFocus( TRUE );

    connect( header(), SIGNAL( clicked(int)),
            SLOT(slotSortingChanged(int) ));

    connect( this, SIGNAL( returnPressed(QListViewItem *) ),
            SLOT( slotActivate( QListViewItem *) ) );

    connect( this, SIGNAL( clicked(QListViewItem *, const QPoint&, int)),
            SLOT( selected( QListViewItem *) ) );
    
    connect( this, SIGNAL( doubleClicked(QListViewItem *, const QPoint&, int)),
            SLOT( slotActivate( QListViewItem *) ) );

    connect( this, SIGNAL(contextMenuRequested( QListViewItem *, const QPoint &, int )),
            this, SLOT( slotActivateMenu( QListViewItem *, const QPoint& )));

    connect( this, SIGNAL( contextMenu (KListView*, QListViewItem*, const QPoint&) ),
            this, SLOT( showContextMenu( KListView*, QListViewItem*, const QPoint& ) ) );

    KFileView::setSelectionMode( KFile::Single );

    // for highlighting
    connect( this, SIGNAL( selectionChanged( QListViewItem * ) ),
            SLOT( highlighted( QListViewItem * ) ));

    setSorting( sorting() );

    m_resolver = new KMimeTypeResolver<PFileViewItem,PFileView>( this );

}

PFileView::~PFileView() {
    CONFIG->setGroup( GENERAL_KEY );
    if ( CONFIG->readBoolEntry( "Save configuration on exit", true ) )
            saveConfig();
    delete m_resolver;
    delete[] cols;
}

void PFileView::RegisterManager( PFileManager *manager_, PFileManager *manager_other_ ) {
    manager = manager_;
    manager_other = manager_other_;

    connect( this, SIGNAL( FileEnter( KFileItem* )),
            manager, SLOT( MakeFile( KFileItem* )));

    connect( this, SIGNAL( CopyFile( const KFileItem* )),
            manager_other, SLOT( CopyFileDlg( const KFileItem* )));

    connect( this, SIGNAL( CopyFiles( const KFileItemList* )),
            manager_other, SLOT( CopyFilesDlg( const KFileItemList* )));
    
    connect( this, SIGNAL( MoveFile( const KFileItem* )),
            manager_other, SLOT( MoveFileDlg( const KFileItem* )));

    connect( this, SIGNAL( MoveFiles( const KFileItemList* )),
            manager_other, SLOT( MoveFilesDlg( const KFileItemList* )));
    
    connect( this, SIGNAL( LinkFile( const KFileItem* )),
            manager_other, SLOT( LinkFileDlg( const KFileItem* )));
    
    connect( this, SIGNAL( LinkFiles( const KFileItemList* )),
            manager_other, SLOT( LinkFilesDlg( const KFileItemList* )));
    
    connect( this, SIGNAL( DeleteFile( const KFileItem* )),
            manager, SLOT( DeleteFileDlg( const KFileItem* )));

    connect( this, SIGNAL( DeleteFiles( const KFileItemList* )),
            manager, SLOT( DeleteFilesDlg( const KFileItemList* )));

    connect( this, SIGNAL( RenameFile( const KFileItem* )),
            manager, SLOT( RenameFile( const KFileItem* )));

    connect( this, SIGNAL( FileProperties( KFileItem* )),
            manager, SLOT( FileProperties( KFileItem* )));
    //------------
    connect( this, SIGNAL( doubleClicked( QListViewItem * )),
            this, SLOT( FileEntered( QListViewItem * )));

    connect( this, SIGNAL( returnPressed( QListViewItem * )),
            this, SLOT( FileEntered( QListViewItem * )));

    KIconLoader *il = kapp->iconLoader();
    drop_menu = new KPopupMenu( this, "droppopupmenu" );
    drop_menu->insertItem( il->loadIconSet( "editcopy", KIcon::Small, 0 ), i18n("Copy"), this, SLOT( copyDroppedFiles() ) );
    drop_menu->insertItem( il->loadIconSet( "move", KIcon::Small, 0 ), i18n("Move"), this, SLOT( moveDroppedFiles() ) );
    drop_menu->insertItem( il->loadIconSet( "link", KIcon::Small, 0 ), i18n("Link"), this, SLOT( linkDroppedFiles() ) );
	
}

bool PFileView::isSizeCol() {
    return d->show_col_size;
}

bool PFileView::isDateCol() {
    return d->show_col_date;
}

bool PFileView::isPermCol() {
    return d->show_col_perm;
}

bool PFileView::isOwnerCol() {
    return d->show_col_owner;
}

bool PFileView::isGroupCol() {
    return d->show_col_group;
}

void PFileView::showSizeCol( bool show ) {
    if (d->show_col_size == show) return;
    if (show) {
	cols[COL_SIZE] = addColumn( i18n( "Size" ) );
	redisplay();
    }
    else {
	removeColumn( cols[COL_SIZE] );
	for( int i = 0; i<NCOLS; i++ )
	    if (cols[i] > cols[COL_SIZE]) cols[i]--;
	cols[COL_SIZE] = NCOLS;
	redisplay();
    }
    d->show_col_size = show;
}

void PFileView::showDateCol( bool show ) {
    if (d->show_col_date == show) return;
    if (show) {
	cols[COL_DATE] = addColumn( i18n( "Date" ) );
	redisplay();
    }
    else {
	removeColumn( cols[COL_DATE] );
	for( int i = 0; i<NCOLS; i++ )
	    if (cols[i] > cols[COL_DATE]) cols[i]--;
	cols[COL_DATE] = NCOLS;
	redisplay();
    }
    d->show_col_date = show;
}

void PFileView::showPermCol( bool show ) {
    if (d->show_col_perm == show) return;
    if (show) {
	cols[COL_PERM] = addColumn( i18n( "Permissions" ) );
	redisplay();
    }
    else {
	removeColumn( cols[COL_PERM] );
	for( int i = 0; i<NCOLS; i++ )
	    if (cols[i] > cols[COL_PERM]) cols[i]--;
	cols[COL_PERM] = NCOLS;
	redisplay();
    }
    d->show_col_perm = show;
}
    
void PFileView::showOwnerCol( bool show ) {
    if (d->show_col_owner == show) return;
    if (show) {
	cols[COL_OWNER] = addColumn( i18n( "Owner" ) );
	redisplay();
    }
    else {
	removeColumn( cols[COL_OWNER] );
	for( int i = 0; i<NCOLS; i++ )
	    if (cols[i] > cols[COL_OWNER]) cols[i]--;
	cols[COL_OWNER] = NCOLS;
	redisplay();
    }
    d->show_col_owner = show;
}
    
void PFileView::showGroupCol( bool show ) {
    if (d->show_col_group == show) return;
    if (show) {
	cols[COL_GROUP] = addColumn( i18n( "Group" ) );
	redisplay();
    }
    else {
	removeColumn( cols[COL_GROUP] );
	for( int i = 0; i<NCOLS; i++ )
	    if (cols[i] > cols[COL_GROUP]) cols[i]--;
	cols[COL_GROUP] = NCOLS;
	redisplay();
    }
    d->show_col_group = show;
}

void PFileView::redisplay() {
    PFileViewItem* item = (PFileViewItem*)firstChild();
    while( item ) {
	item->init();
        item = (PFileViewItem*)item->nextSibling();
    }
}

QString PFileView::sortingKey( const QString& value, bool isDir, int sortSpec ) {
    bool reverse   = sortSpec & QDir::Reversed;
    bool dirsFirst = sortSpec & QDir::DirsFirst;
    char start = (isDir && dirsFirst) ? (reverse ? '3' : '1') : '2';
    QString result = (sortSpec & QDir::IgnoreCase) ? value.lower() : value;
    return result.prepend( start );
}

QString PFileView::sortingKey( KIO::filesize_t value, bool isDir, int sortSpec) {
    bool reverse = sortSpec & QDir::Reversed;
    bool dirsFirst = sortSpec & QDir::DirsFirst;
    char start = (isDir && dirsFirst) ? (reverse ? '3' : '1') : '2';
    return KIO::number( value ).rightJustify( 24, '0' ).prepend( start );
}

void PFileView::FileEntered( QListViewItem *item ) {
    if (item)
            emit FileEnter( ((PFileViewItem*)item)->fileInfo() );
}

void PFileView::FileEntered() {
    if ( currentFileItem() )
            emit FileEnter( currentFileItem() );
}

void PFileView::CopyFiles() {
    const KFileItemList* list = KFileView::selectedItems();
    if (!list) return;
    if ( list->isEmpty() ) {
            const KFileItem* current = currentFileItem();
            if (!current || current->name() == ".." ) return;
            else emit CopyFile( current );
    }
    else if ( list->count() == 1 )
            emit CopyFile( list->getFirst() );
    else
            emit CopyFiles( list );
}

void PFileView::MoveFiles() {
    const KFileItemList* list = KFileView::selectedItems();
    if (!list) return;
    if ( list->isEmpty() ) {
            const KFileItem* current = currentFileItem();
            if (!current || current->name() == ".." ) return;
            else emit MoveFile( current );
    }
    else if ( list->count() == 1 )
            emit MoveFile( list->getFirst() );
    else
            emit MoveFiles( list );
}
void PFileView::DeleteFiles() {
    const KFileItemList* list = KFileView::selectedItems();
    if (!list) return;
    if ( list->isEmpty() ) {
            const KFileItem* current = currentFileItem();
            if (!current || current->name() == ".." ) return;
            else emit DeleteFile( current );
    }
    else if ( list->count() == 1 )
            emit DeleteFile( list->getFirst() );
    else
            emit DeleteFiles( list );
}

void PFileView::RenameFile() {
    KFileItem *file = currentFileItem();
    if (file && file->name() != ".." )
            emit RenameFile( file );
}

void PFileView::FileProperties() {
    KFileItem *file = currentFileItem();
    if ( file && file->name() != ".." )
            emit FileProperties( file );
}

void PFileView::LinkFiles() {
    const KFileItemList* list = KFileView::selectedItems();
    if (!list) return;
    if ( list->isEmpty() ) {
            const KFileItem* current = currentFileItem();
            if (!current || current->name() == ".." ) return;
            else emit LinkFile( current );
    }
    else if ( list->count() == 1 )
            emit LinkFile( list->getFirst() );
    else
            emit LinkFiles( list );
}

void PFileView::CopyFilesClip() {
kapp->clipboard()->setData( dragObject() );
}

void PFileView::CutFiles() {
}

void PFileView::PasteFiles() {
    d->dropped_to_url = manager->CurrentDirectory();
    d->files_dropped_list.clear();
    QUriDrag::decodeToUnicodeUris( kapp->clipboard()->data(), d->files_dropped_list );
    copyDroppedFiles();
}

KFileItem *PFileView::setSelected( const QString& info, bool enable ) {
    if ( info.isNull() )
            return 0;

    KFileItemListIterator it( *items() );
    for( ; it.current() ; ++it )
            if ( info == (*it)->name() ) {
                    setSelected( *it, enable );
                    return *it;
            }
    return 0;
}

void PFileView::setSelected( const KFileItem *info, bool enable ) {
    if ( !info )
            return;

    // we can only hope that this casts works
    PFileViewItem *item = (PFileViewItem*)info->extraData( this );

    if ( item )
            item->setmSelected( enable );
}

KFileItem *PFileView::setCurrentItem( const QString &filename ) {
    if (!filename.isNull()) {
            KFileItem *item;
            for ( (item = firstFileItem()); item; item = nextItem( item ) ) {
                    if (item->name() == filename) {
                            setCurrentItem( item );
                            return item;
                    }
            }
    }
    return 0;
}

void PFileView::setCurrentItem( const KFileItem *item ) {
    if ( !item )
            return;
    
    PFileViewItem *it = (PFileViewItem*) item->extraData( this );
    if ( it )
            PListView::setCurrentItem( it );
}

KFileItem * PFileView::currentFileItem() const {
    PFileViewItem *current = static_cast<PFileViewItem*>( currentItem() );
    if ( current )
            return current->fileInfo();

    return 0L;
}

void PFileView::clearSelection() {
    PListView::clearSelection();
}

void PFileView::selectAll() {
    if (KFileView::selectionMode() == KFile::NoSelection ||
    KFileView::selectionMode() == KFile::Single)
            return;

    PListView::mselectAll( true );
}

void PFileView::invertSelection() {
    PListView::invertSelection();
}

void PFileView::slotActivateMenu (QListViewItem *item,const QPoint& pos ) {
    if ( !item ) {
            sig->activateMenu( 0, pos );
            return;
    }
    PFileViewItem *i = (PFileViewItem*) item;
    sig->activateMenu( i->fileInfo(), pos );
}

void PFileView::showFileProperties( QListViewItem *item ) {
    if (item)
            emit FileProperties( ((PFileViewItem*)item)->fileInfo() );
}

void PFileView::showContextMenu( KListView *, QListViewItem *item, const QPoint& p) {
    if (item) {
            KFileItem *file = ((PFileViewItem*)item)->fileInfo();
            if (file && file->name() != ".." ) {
                    updateFileMenu( file );
                    Pcmm::file_menu->popup( p );
            }
    }
    else
            Pcmm::operations_menu->popup( p );
}

void PFileView::filesDropped( QDropEvent *event) {
    QListViewItem *atpos = itemAt(contentsToViewport(event->pos()));
    if (atpos) {
            KFileItem *file = ((PFileViewItem*)atpos)->fileInfo();
            d->dropped_to_url = file->url();
    }
    else {
            d->dropped_to_url = manager->CurrentDirectory();
    }
    QByteArray ba;
    ba = event->encodedData( event->format() );
    d->files_dropped_list = QStringList::split( "\r\n", KURL::decode_string( ba, UTF8_MIB ) );
    drop_menu->popup( mapToGlobal(contentsToViewport(event->pos() ) ) );
}

void PFileView::copyDroppedFiles() {
    if (d->files_dropped_list.isEmpty() )
            return;

    if (d->files_dropped_list.count() == 1)
            manager->CopyFile( d->files_dropped_list.first(), d->dropped_to_url );
    else
            manager->CopyFiles( d->files_dropped_list, d->dropped_to_url );
}

void PFileView::moveDroppedFiles() {
    if (d->files_dropped_list.isEmpty() )
            return;

    if (d->files_dropped_list.count() == 1)
            manager->MoveFile( d->files_dropped_list.first(), d->dropped_to_url );
    else
            manager->MoveFiles( d->files_dropped_list, d->dropped_to_url );
}

void PFileView::linkDroppedFiles() {
    if (d->files_dropped_list.isEmpty() )
            return;

    if (d->files_dropped_list.count() == 1)
            manager->LinkFile( d->files_dropped_list.first(), d->dropped_to_url );
    else
            manager->LinkFiles( d->files_dropped_list, d->dropped_to_url );
}

void PFileView::openWithActivated( int id ) {
    KFileItem *file = currentFileItem();
    int index = d->open_popup->indexOf( id );
    if (!file) return;
    if (index > ((int)d->file_offers.count())-1)
        SLOTS->OpenWith();
    else
        KRun::run( *d->file_offers[index], KURL::List(file->url() ) );
}

void PFileView::clearView() {
	m_resolver->m_lstPendingMimeIconItems.clear();
	PListView::clear();
}

void PFileView::readConfig() {
    KConfig *config = CONFIG;
    config->setGroup(d->is_left_panel ? LEFT_PANEL_KEY : RIGHT_PANEL_KEY );
    cols[COL_NAME] = addColumn( i18n( "Name" ) );
    if (d->show_col_size = config->readBoolEntry( "Col SIZE", true ) )
	cols[COL_SIZE] = addColumn( i18n( "Size" ) );
    else
	cols[COL_SIZE] = -1;
    if (d->show_col_date = config->readBoolEntry( "Col DATE", true ) )
	cols[COL_DATE] = addColumn( i18n( "Date" ) );
    else
	cols[COL_DATE] = -1;
    if (d->show_col_perm = config->readBoolEntry( "Col PERM", false ) )
	cols[COL_PERM] = addColumn( i18n( "Permissions" ) );
    else
	cols[COL_PERM] = -1;
    if (d->show_col_owner = config->readBoolEntry( "Col OWNER", true ) )
	cols[COL_OWNER] = addColumn( i18n( "Owner" ) );
    else
	cols[COL_OWNER] = -1;
    if (d->show_col_group = config->readBoolEntry( "Col GROUP", true ) )
	cols[COL_GROUP] = addColumn( i18n( "Group" ) );
    else
	cols[COL_GROUP] = -1;
}

void PFileView::saveConfig() {
    KConfig *config = CONFIG;
    config->setGroup( d->is_left_panel ? LEFT_PANEL_KEY : RIGHT_PANEL_KEY );
    config->writeEntry( "Col SIZE", d->show_col_size );
    config->writeEntry( "Col DATE", d->show_col_date );
    config->writeEntry( "Col PERM", d->show_col_perm );
    config->writeEntry( "Col OWNER", d->show_col_owner );
    config->writeEntry( "Col GROUP", d->show_col_group );
    config->sync();
}

void PFileView::insertItem( KFileItem *i ) {
    KFileView::insertItem( i );

    PFileViewItem *item = new PFileViewItem( (QListView*) this, i );

    setSortingKey( item, i );

    i->setExtraData( this, item );

    if ( !i->isMimeTypeKnown() )
            m_resolver->m_lstPendingMimeIconItems.append( item );
}

void PFileView::slotActivate( QListViewItem *item ) {
    if ( !item )
            return;

    const KFileItem *fi = ( (PFileViewItem*)item )->fileInfo();
    if ( fi )
            sig->activate( fi );
}

void PFileView::selected( QListViewItem *item ) {
    if ( !item )
            return;

    if ( KGlobalSettings::singleClick() ) {
            const KFileItem *fi = ( (PFileViewItem*)item )->fileInfo();
    if ( fi && (fi->isDir() || !onlyDoubleClickSelectsFiles()) )
            sig->activate( fi );
    }
}

void PFileView::highlighted( QListViewItem *item ) {
    if ( !item )
            return;

    const KFileItem *fi = ( (PFileViewItem*)item )->fileInfo();
    if ( fi )
            sig->highlightFile( fi );
}

bool PFileView::isSelected( const KFileItem *i ) const {
    if ( !i )
            return false;

    PFileViewItem *item = (PFileViewItem*) i->extraData( this );
    return (item && item->ismSelected());
}


void PFileView::updateView( bool b ) {
    if ( !b )
            return;

    QListViewItemIterator it( (QListView*)this );
    for ( ; it.current(); ++it ) {
            PFileViewItem *item=static_cast<PFileViewItem *>(it.current());
            item->setPixmap( 0, item->fileInfo()->pixmap(KIcon::SizeSmall) );
    }
}

void PFileView::updateView( const KFileItem *i ) {
    if ( !i )
            return;

    PFileViewItem *item = (PFileViewItem*) i->extraData( this );
    if ( !item )
            return;

    item->init();
    setSortingKey( item, i );

    //item->repaint(); // only repaints if visible
}

void PFileView::setSortingKey( PFileViewItem *item, const KFileItem *i ) {
    // see also setSorting()
    QDir::SortSpec spec = KFileView::sorting();
    bool reverse = spec & QDir::Reversed;
    char c = reverse ? '4' : '0';

    if ( spec & QDir::Time ) {
        QString key = sortingKey( i->time( KIO::UDS_MODIFICATION_TIME ), i->isDir(), spec ); 
        if (i->name() == ".." ) key[0] = c;
        item->setKey( key );
    }
    else if ( spec & QDir::Size ) {
        QString key = sortingKey( i->size(), i->isDir(), spec );
        if (i->name() == ".." ) key[0] = c;
        item->setKey( key );
    }

    else {// Name or Unsorted
        QString key = sortingKey( i->text(), i->isDir(), spec );
        if (i->name() == ".." ) key[0] = c;
        item->setKey( key );
    }
}


void PFileView::removeItem( const KFileItem *i ) {
    if ( !i )
            return;

    PFileViewItem *item = (PFileViewItem*) i->extraData( this );
    m_resolver->m_lstPendingMimeIconItems.remove( item );
    delete item;

    KFileView::removeItem( i );
    delete i;
}

KFileItem *PFileView::removeItem( const QString& i ) {
    if ( i.isNull() )
            return 0;

    KFileItemListIterator it( *items() );
    for( ; it.current() ; ++it )
            if ( i == (*it)->name() ) {
                    removeItem( *it );
                    return *it;
            }
    return 0;
}

void PFileView::slotSortingChanged( int col ) {
    QDir::SortSpec sort = sorting();
    int sortSpec = -1;
    bool reversed = col == m_sortingCol && (sort & QDir::Reversed) == 0;
    m_sortingCol = col;

    if ( col == cols[COL_NAME] )
        sortSpec = (sort & ~QDir::SortByMask | QDir::Name);
    else if ( col == cols[COL_SIZE] )
        sortSpec = (sort & ~QDir::SortByMask | QDir::Size);
    else if ( col == cols[COL_DATE] )
        sortSpec = (sort & ~QDir::SortByMask | QDir::Time);
    // the following columns have no equivalent in QDir, so we set it
    // to QDir::Unsorted and remember the column (m_sortingCol)
    else if ( col == cols[COL_OWNER] || col == cols[COL_GROUP] || col == cols[COL_PERM] )
        // grmbl, QDir::Unsorted == SortByMask.
        sortSpec = (sort & ~QDir::SortByMask);// | QDir::Unsorted;

    if ( reversed )
            sortSpec |= QDir::Reversed;
    else
            sortSpec &= ~QDir::Reversed;

    if ( sort & QDir::IgnoreCase )
            sortSpec |= QDir::IgnoreCase;
    else
            sortSpec &= ~QDir::IgnoreCase;


    KFileView::setSorting( static_cast<QDir::SortSpec>( sortSpec ) );

    KFileItem *item;
    KFileItemListIterator it( *items() );

    char c = reversed ? '4' : '0';

    if ( sortSpec & QDir::Time ) {
            for ( ; (item = it.current()); ++it ) {
                QString key = sortingKey( item->time( KIO::UDS_MODIFICATION_TIME ), item->isDir(), sortSpec );
                if (item->name() == ".." ) key[0] = c;
                viewItem(item)->setKey( key );
            }
    }

    else if ( sortSpec & QDir::Size ) {
            for ( ; (item = it.current()); ++it ) {
                QString key = sortingKey( item->size(), item->isDir(), sortSpec );
                if (item->name() == ".." ) key[0] = c;
                viewItem(item)->setKey( key );
            }
    }
    else { // Name or Unsorted -> use column text
            for ( ; (item = it.current()); ++it ) {
                    PFileViewItem *i = viewItem( item );
                    QString key = sortingKey( i->text(m_sortingCol), item->isDir(), sortSpec );
                    if (item->name() == ".." ) key[0] = c;
                    i->setKey( key );
            }
    }

    PListView::setSorting( m_sortingCol, !reversed );
    PListView::sort();

    if ( !d->m_blockSortingSignal )
            sig->changeSorting( static_cast<QDir::SortSpec>( sortSpec ) );
}


void PFileView::setSorting( QDir::SortSpec spec ) {
    int col = 0;
    if ( spec & QDir::Time )
            col = cols[COL_DATE];
    else if ( spec & QDir::Size )
            col = cols[COL_SIZE];
    else if ( spec & QDir::Unsorted )
            col = m_sortingCol;
    else
            col = cols[COL_NAME];

    // inversed, because slotSortingChanged will reverse it
    if ( spec & QDir::Reversed )
            spec = (QDir::SortSpec) (spec & ~QDir::Reversed);
    else
            spec = (QDir::SortSpec) (spec | QDir::Reversed);

    m_sortingCol = col;
    KFileView::setSorting( (QDir::SortSpec) spec );


    // don't emit sortingChanged() when called via setSorting()
    d->m_blockSortingSignal = true; // can't use blockSignals()
    slotSortingChanged( col );
    d->m_blockSortingSignal = false;
}


void PFileView::ensureItemVisible( const KFileItem *i ) {
    if ( !i )
            return;
    PFileViewItem *item = (PFileViewItem*) i->extraData( this );
    if ( item )
            PListView::ensureItemVisible( item );
}

KFileItem * PFileView::firstFileItem() const {
PFileViewItem *item = static_cast<PFileViewItem*>( firstChild() );
if ( item )
        return item->fileInfo();
return 0L;
}

KFileItem * PFileView::nextItem( const KFileItem *fileItem ) const {
    if ( fileItem ) {
            PFileViewItem *item = viewItem( fileItem );
            if ( item && item->itemBelow() )
                    return ((PFileViewItem*) item->itemBelow())->fileInfo();
            else
                    return 0L;
    }
    else
            return firstFileItem();
}

KFileItem * PFileView::prevItem( const KFileItem *fileItem ) const {
    if ( fileItem ) {
            PFileViewItem *item = viewItem( fileItem );
            if ( item && item->itemAbove() )
                    return ((PFileViewItem*) item->itemAbove())->fileInfo();
            else
                    return 0L;
    }
    else
            return firstFileItem();
}

void PFileView::keyPressEvent( QKeyEvent *e ) {
    PListView::keyPressEvent( e );

    if ( e->key() == Key_Return || e->key() == Key_Enter ) {
            if ( e->state() & ControlButton )
                    e->ignore();
    else
            e->accept();
    }
}

void PFileView::focusInEvent( QFocusEvent *event ) {
    PListView::focusInEvent( event );
    ((PFilePanel*)parent())->focusInEventChild();
}

void PFileView::focusOutParent() {
if ( currentItem() ) {
    currentItem()->setSelected(false);
    currentItem()->repaint();
}
}

void PFileView::startDrag() {
    QDragObject *drag = dragObject();

    if (!drag)
          return;

    drag->drag();
}

QDragObject *PFileView::dragObject() {
    const KFileItemList *file_list = KFileView::selectedItems();
    KFileItem *file = currentFileItem();
    if ( file_list->isEmpty() && ( !file || file->name() == ".." ) )
            return 0;

    QStrList uri_list;
    if ( !file_list->isEmpty() ) {
            KFileItemListIterator it( *file_list );
            for ( ; it.current() ; ++it )
                    if ( isSelected( *it ) ) {
                            uri_list.append( (*it)->url().url( 0, UTF8_MIB ) );
                    }
    }
    else { //aktualni soubor
            uri_list.append( file->url().url( 0, UTF8_MIB ) );
    }
    return new QUriDrag( uri_list, this );
}

bool PFileView::acceptDrag (QDropEvent* event) const {
    QListViewItem *item = itemAt(contentsToViewport(event->pos()));
    bool m_accept = true;
    if (item) {
            KFileItem *file = ((PFileViewItem*)item)->fileInfo();
            m_accept = file->isDir();
    }

    return m_accept && event->provides( "text/uri-list" );
}

//
// mimetype determination on demand
//
void PFileView::mimeTypeDeterminationFinished() {
// anything to do?
}

void PFileView::determineIcon( PFileViewItem *item ) {
    (void) item->fileInfo()->determineMimeType();
    updateView( item->fileInfo() );
}

void PFileView::listingCompleted() {
    m_resolver->start();
}

void PFileView::updateFileMenu( KFileItem *file ) {
    KPopupMenu *file_menu = Pcmm::file_menu;
    file_menu->changeTitle( Pcmm::file_menu_title_id, file->name() );
    int open_with_index = file_menu->indexOf( Pcmm::file_menu_open_with_id );
    file_menu->removeItem( Pcmm::file_menu_open_with_id );

    if (file->isDir()) {
        Pcmm::file_menu_open_with_id = file_menu->insertItem(
                i18n("Open with..."), SLOTS, SLOT( OpenWith() ), 0, -1, open_with_index );
    }
    else {
        d->open_popup = new KPopupMenu(this);
        d->file_offers = KTrader::self()->query( file->mimetype() );
        KTrader::OfferList::iterator it;
        for (it = d->file_offers.begin(); it != d->file_offers.end(); ++it )
            d->open_popup->insertItem( (*it)->pixmap( KIcon::Toolbar ), (*it)->name() );
        d->open_popup->insertSeparator();
        d->open_popup->insertItem( i18n("Others...") );
        Pcmm::file_menu_open_with_id = file_menu->insertItem( i18n("Open with"), d->open_popup, -1, open_with_index );
        connect( d->open_popup, SIGNAL( activated( int ) ), SLOT( openWithActivated( int ) ) );
    }
}

/////////////////////////////////////////////////////////////////

QPixmap *PFileViewItem::up_icon = 0;

void PFileViewItem::init() {
    PFileView *view = (PFileView*)listView();
    if( inf->text() == ".." ) {
        PFileViewItem::setPixmap( 0, *up_icon );
        setText( view->cols[COL_NAME], ".." );
    }
    else {
        PFileViewItem::setPixmap( 0, inf->pixmap(KIcon::SizeSmall));

        setText( view->cols[COL_NAME], inf->text() );
        setText( view->cols[COL_SIZE], fileInfo()->isDir() ? QString("<DIR>") : KGlobal::locale()->formatNumber( inf->size(), 0));
        setText( view->cols[COL_DATE],  inf->timeString() );
        setText( view->cols[COL_PERM],  inf->permissionsString() );
        setText( view->cols[COL_OWNER], inf->user() );
        setText( view->cols[COL_GROUP], inf->group() );
    }
}

void PFileViewItem::setmSelected( bool sel ) {
    if ( inf->name() != ".." )
            PListViewItem::setmSelected( sel );
}

void PFileView::virtual_hook( int id, void* data ) {
    PListView::virtual_hook( id, data );
    KFileView::virtual_hook( id, data );
}
