#include <qfile.h>
#include <qdir.h>
#include <qtextstream.h>
#include <qtextbrowser.h>
#include <qapplication.h>
#include <qwidget.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qlayout.h>
#include <qvaluelist.h>
// KDE includes
#include <kmimetype.h>
#include <kio/netaccess.h>
#include <kparts/part.h>
#include <klibloader.h>
#include <ktrader.h>
#include <kcursor.h>
#include <kmessagebox.h>
#include <kprocess.h>

#include "pviewer.h"
#include "pviewerwidget.h"
#include "../pcmm_general.h"


void PViewer::view(QWidget *parent, QString fileName, QString mimetype) {
    parent->setCursor(KCursor::waitCursor());

    KURL url = fileName;
    QString tmpFile = QString::null;

    // the following part is for local files
    if (url.isLocalFile()) {
        fileName = url.path();
        // translate symlink to real names
        for( QFileInfo qf(fileName); qf.isSymLink(); qf.setFile(fileName) ){
            fileName = qf.dir().absFilePath(qf.readLink());
        }
        if( !QDir().exists(fileName) ) return;  // safty
        tmpFile = fileName;
    } 
    else {
        // files that are not local must first be downloaded
        // unless for html files that is : )
        if( mimetype.left(9)!="text/html" ){
            if( !KIO::NetAccess::download( url, tmpFile ) ){
                KMessageBox::sorry(parent,i18n("PViewer is unable to download: ")+fileName);
                return;
            }
        }
        else tmpFile = fileName;
    }
    url.setPath(tmpFile);
    mimetype = KMimeType::findByURL(url,0,true,false)->name();

    short Type=0;                            // the default is a KPART viewer
    if ( mimetype.left(5)=="image") Type=1;  // image
    if ( mimetype.contains("text") &&
        !mimetype.contains("html")) Type=2;  // text
    if ( mimetype.contains("executable") ||
        mimetype.contains("object-file")||
        mimetype.contains("octet-stream") ) Type =3; // binary format

    switch (Type) {
        case 0 : viewGeneric(parent,tmpFile,mimetype);           break;
        case 1 : viewImage(parent,tmpFile);                      break;
        case 2 : viewText(parent,tmpFile);                       break;
        case 3 : viewHex(parent,tmpFile);                        break;
    }
    // delete temp file ( if there is one )
    if( tmpFile != fileName ) KIO::NetAccess::removeTempFile( tmpFile );
    parent->setCursor(KCursor::arrowCursor());
}

void PViewer::viewImage( QWidget *parent, QString fileName) {
    // first, open and load the image
    QPixmap image;
    if (!image.load(fileName)) {
        KMessageBox::sorry(parent,i18n("PViewer is unable to load the image."));
        return;
    }
    // now, to the widget...	
    PViewerWidget *dock=new PViewerWidget(parent);
    // check if the picture is bigger than the screen
    short sw=QApplication::desktop()->width();
    short sh=QApplication::desktop()->height();
    if (image.width() > sw || image.height() > sh-90) {
        // perspectively-scale down the image to a viewable size
        short newHeight=sh-90; // allow space for a large-panel
        short newWidth=image.width()*newHeight/image.height(); // perspective
        QImage temp=image.convertToImage();
        temp=temp.smoothScale(newWidth,newHeight);
        if (!image.convertFromImage(temp,QPixmap::AutoColor)) {
            KMessageBox::sorry(parent,i18n("PViewer is unable to load the image."));
            return;
        }
    }
    dock->resize(image.width(), image.height());
    dock->setBackgroundPixmap(image);
    dock->setCaption("(image) PViewer: "+fileName);
    
    // let's go.
    dock->exec();	
}

void PViewer::viewGeneric( QWidget *parent, QString fileName, QString mimetype, QString caption) {
    KParts::ReadOnlyPart *kviewer_part = 0L;
    KLibFactory *factory = 0;
    KTrader::OfferList offers = KTrader::self()->query(mimetype.latin1());

    PViewerWidget *viewer = new PViewerWidget(parent);
    QVBoxLayout* layout=new QVBoxLayout(viewer);

    // in theory, we only care about the first one.. but let's try all
    // offers just in case the first can't be loaded for some reason
    KTrader::OfferList::Iterator it(offers.begin());
    for( ; it != offers.end(); ++it) {
        KService::Ptr ptr = (*it);
        // we now know that our offer can handle mimetype and is a part.
        // since it is a part, it must also have a library... let's try to
        // load that now
        factory = KLibLoader::self()->factory( ptr->library().latin1() );
        if (factory) {
            kviewer_part = static_cast<KParts::ReadOnlyPart *>(factory->create(viewer,
                                 ptr->name().latin1(), "KParts::ReadOnlyPart"));
            if( kviewer_part && kviewer_part->openURL(fileName) ) break;
            else {
                delete kviewer_part;
                kviewer_part = 0L;
            }
        }
    }

    if (!kviewer_part){
        delete viewer;  // also delete the layout
        viewSpecial(parent,fileName);
        return;
    }

    layout->addWidget(kviewer_part->widget());
    // let's go.
    if (caption.isEmpty()) caption = fileName;
    viewer->setCaption("PViewer: "+caption);
    viewer->resize(parent->width()*2/3,parent->height()*2/3);	
    viewer->exec();
}

// this function is called if the file isn't an html or an image, so
// a simple editor is used
void PViewer::viewText( QWidget *parent, QString fileName) {
    // first, create a stream to the file
    QFile file(fileName);
    file.open(IO_ReadOnly);
    // now, to the widget...	
    PViewerWidget *dock = new PViewerWidget(parent);
    PEdit *viewer = new PEdit(dock);
    connect(dock,SIGNAL(search()),viewer,SLOT(search()));
    QVBoxLayout* layout=new QVBoxLayout(dock); // <patch> thanks to Aurelien Gateau
    viewer->setReadOnly(true);

    // since we want a QUICK view - we read only the first 5000 lines.
    // we also don't read more than 200 chars per line - lines longer than that
    // are probably part of a binary file...
    QString line;
    for(int i = 0; i < 5000 ; ++i) {
        if ( file.readLine(line,200)== -1 ) break;
        else {
            if(line.right(1) == "\n" ) line.truncate(line.length()-1);
            viewer->insertLine(line);
        }
    }

    //if we cut the file short - notify the user.
    if(file.readLine(line,10) != -1){
        viewer->insertLine("\n********* PCMM MESSAGE ********");
        viewer->insertLine("ONLY THE FIRST 5000 LINES ARE SHOWN:");
        viewer->insertLine("USE F4 TO SEE THE WHOLE FILE.");
    }

    file.close();

    // let's go.
    layout->addWidget(viewer);
    dock->setCaption("PViewer: "+fileName);
    dock->resize(parent->width()*2/3,parent->height()*2/3);	
    dock->exec();
}

void PViewer::viewSpecial( QWidget *parent, QString fileName){
    // get the file info from the 'file' command.
    QString tmpFile = getTempFile();
    KShellProcess proc;
    proc << "file"<< "\""+fileName+"\"" <<" > "+tmpFile ;
    proc.start(KProcess::Block);

    // create a stream to the file
    QFile file(tmpFile);
    file.open(IO_ReadOnly);
    QString line;
    QString fileInfo;
    while( file.readLine(line,200) != -1 ){
        if(line.right(1) == "\n" ) line.truncate(line.length()-1);
        fileInfo += line + '\n';
    }
    QDir().remove(tmpFile);
//    kdWarning() << fileInfo << endl;

    int res = KMessageBox::warningYesNoCancel(0, i18n(
    "Pcmm can't find a viewer for %1 How do you want to view it? Warning: Pcmm may freeze or crash if you'll try to view a special file!").
        arg(fileInfo), i18n("Unknown viewer"), i18n("&Text view"), i18n("&Hex dump"));
    if (res == KMessageBox::Yes) viewText(parent,fileName);
    if (res == KMessageBox::No) viewHex(parent,fileName);
}

void PViewer::viewHex( QWidget *parent, QString fileName){
    QString tmpFile = getTempFile();
    KShellProcess proc;
    proc << "hexdump -v "<< "\""+fileName+"\"" <<" > "+tmpFile ;
    proc.start(KProcess::Block);

    viewGeneric(parent,tmpFile,"text/plain",fileName);
    QDir().remove(tmpFile);
}

QString PViewer::getTempFile() {
    // try to make temp file
    kapp->config()->setGroup( GENERAL_KEY );
    QString tmpDir = kapp->config()->readEntry( "Temporary directory", "/tmp/pcmm.tmp" );
    for(int i = 1 ; i != -1 ; i=tmpDir.find('/',i+1) )
    QDir().mkdir( tmpDir.left(i) );
    QDir().mkdir( tmpDir );
    chmod( tmpDir.local8Bit(), 0777);
    // add a secure sub dir under the user UID
    QString uid;
    uid.sprintf("%d",getuid());
    QDir(tmpDir).mkdir(uid);
    tmpDir=tmpDir+"/"+uid+"/";
    chmod(tmpDir.local8Bit(),S_IRUSR | S_IWUSR | S_IXUSR );

    while(QDir().exists(tmpDir)) tmpDir = tmpDir+kapp->randomString(8);
    return tmpDir;
}
